from unittest.mock import MagicMock

import dill
from pika import BlockingConnection
from pika.adapters.blocking_connection import BlockingChannel

from .client import Client


def test_init():
    conn = MagicMock(
        spec=BlockingConnection, add_callback_threadsafe=lambda func: func()
    )
    ch = MagicMock(
        spec=BlockingChannel,
        basic_publish=lambda body, **kwargs: client._waiting[
            dill.loads(body)["key"]
        ].set_result("heyo"),
    )
    client = Client("", conn, ch)

    assert client.hello_world("hi") == 'heyo'



from pathlib import Path

from poetry_publish.publish import poetry_publish

import mause_rpc


def publish():
    poetry_publish(
        package_root=Path(mause_rpc.__file__).parent.parent,
        version=mause_rpc.__version__,
    )



from concurrent.futures import Future
from dataclasses import dataclass, field
from functools import partial
from threading import Thread
from typing import Dict
from uuid import uuid4

import dill
from pika import BasicProperties, BlockingConnection
from pika.adapters.blocking_connection import BlockingChannel

from .config import get_parameters


@dataclass
class Client:
    server_queue: str
    conn: BlockingConnection
    channel: BlockingChannel
    _waiting: Dict[str, Future] = field(default_factory=dict)

    def __post_init__(self):
        self.channel.basic_consume("amq.rabbitmq.reply-to", self.recieve, auto_ack=True)

    def call(self, method, *args, **kwargs):
        f: Future = Future()
        key = uuid4().hex
        self._waiting[key] = f
        self.conn.add_callback_threadsafe(
            lambda: self.channel.basic_publish(
                exchange="",
                routing_key=self.server_queue,
                body=dill.dumps(
                    {"method": method, "args": args, "kwargs": kwargs, "key": key}
                ),
                properties=BasicProperties(reply_to="amq.rabbitmq.reply-to"),
            )
        )
        return f.result(timeout=10)

    def __getattr__(self, method):
        return partial(self.call, method)

    def recieve(self, ch, method_frame, properties, body):
        body = dill.loads(body)
        f = self._waiting.pop(body["key"])
        if "body" in body:
            f.set_result(body["body"])
        elif "exception" in body:
            f.set_exception(body["exception"])
        else:
            raise RuntimeError()


def get_client(server_queue: str, server_connection_parameters) -> Client:
    conn = BlockingConnection(server_connection_parameters)
    channel = conn.channel()

    t = Thread(target=channel.start_consuming)
    t.daemon = True
    t.start()

    return Client(server_queue, conn, channel)



import os
from functools import lru_cache
from urllib.parse import urlparse

import pika


@lru_cache()
def get_parameters():
    url_str = os.environ['CLOUDAMQP_URL']
    url = urlparse(url_str)
    return pika.ConnectionParameters(
        host=url.hostname,
        virtual_host=url.path[1:],
        credentials=pika.PlainCredentials(url.username, url.password),
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mause_rpc']

package_data = \
{'': ['*']}

install_requires = \
['dill>=0.3.1,<0.4.0', 'pika>=1.1.0,<2.0.0', 'retry>=0.9.2,<0.10.0']

entry_points = \
{'console_scripts': ['publish = mause_rpc:publish']}

setup_kwargs = {
    'name': 'mause-rpc',
    'version': '0.0.1',
    'description': '',
    'long_description': None,
    'author': 'Elliana May',
    'author_email': 'me@mause.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
import socket
from dataclasses import dataclass, field
from typing import Callable, Dict

import dill
import pika
from pika import ConnectionParameters
from retry import retry

from .config import get_parameters

logging.basicConfig(level=logging.INFO)
logging.getLogger("pika").setLevel(logging.WARN)


@dataclass
class Server():
    server_queue: str
    connection_params: ConnectionParameters
    _methods: Dict[str, Callable] = field(default_factory=dict)

    def register(self, method: Callable):
        self._methods[method.__name__] = method
        return method

    @retry(socket.gaierror, delay=10, jitter=(1, 3))
    @retry(pika.exceptions.ChannelClosedByBroker, delay=10, jitter=(1, 3))
    @retry(pika.exceptions.AMQPConnectionError, delay=5, jitter=(1, 3))
    def serve(self):
        with pika.BlockingConnection(self.connection_params) as conn:
            channel = conn.channel()

            channel.queue_declare(queue=self.server_queue, exclusive=True, auto_delete=True)
            channel.basic_consume(self.server_queue, self.on_server_rx_rpc_request)
            logging.info("Ready")
            channel.start_consuming()

    def on_server_rx_rpc_request(self, ch, method_frame, properties, body):
        body = dill.loads(body)
        logging.info("RPC Server got request: %s", body)

        res = {"key": body["key"]}

        try:
            res["body"] = self._methods[body["method"]](*body["args"], **body["kwargs"])
        except Exception as e:
            logging.exception("Call to %s caused exception", body["method"], e)
            res["exception"] = e

        ch.basic_publish("", routing_key=properties.reply_to, body=dill.dumps(res))

        ch.basic_ack(delivery_tag=method_frame.delivery_tag)
